@@ -8,13 +8,10 @@ import subprocess
 import json
 import yaml
 import shutil
-import click
-import sys
-from pprint import pprint
+import threading
 from datetime import datetime
 from github import Github
 from botocore.exceptions import WaiterError
-
 logging.basicConfig(level=logging.INFO)
 
 
@@ -26,11 +23,10 @@ class EC2Manager:
         self._github_token = os.environ.get('GITHUB_TOKEN', kwargs.get('github_token'))
 
         # optional envs
-        self._config_file = os.environ.get('EC2_MANAGER_CONFIG', kwargs.get('config'))
+        self._config_file = os.environ.get('EC2_MANAGER_CONFIG', kwargs.get('config', 'config.yaml'))
         self._commit = os.environ.get('COMMIT', 'master')
-        self._aws_backend_bucket = os.environ.get('AWS_BACKEND_BUCKET', 'ec2-manager-terraform-state')
         self._terraform_directory = os.environ.get(
-            'TERRAFORM_DIRECTORY', os.path.join(os.getcwd(), 'ec2_manager', 'terraform')
+            'TERRAFORM_DIRECTORY', os.path.join(os.getcwd(), 'terraform')
         )
         self._template_directory = os.path.join(os.path.dirname(__file__), 'template')
         self._repo_url = (
@@ -38,8 +34,13 @@ class EC2Manager:
         )
         self._max_attempts = int(os.environ.get('MAX_TIMEOUT', 600))
 
+        self._github_client = Github(self._github_token)
+
+        self._logger = logging.getLogger(self.__class__.__name__)
+        self._logger.setLevel(level=logging.INFO)
+
         # only when initialize with a config
-        if kwargs.get('config'):
+        if self._config_file:
             # internal variables
             self._config = self._get_config()
             self._aws_default_region = self._config['aws_region']
@@ -55,7 +56,15 @@ class EC2Manager:
             self._ssm_client = self._aws_session.client('ssm')
             self._ec2_client = self._aws_session.client('ec2')
             self._s3_client = self._aws_session.client('s3')
-            self._github_client = Github(self._github_token)
+
+            account_number = self._aws_session.client('sts').get_caller_identity().get('Account')
+            if not account_number:
+                raise ValueError(
+                    'Please set the name of your terraform backend bucket with the AWS_BACKEND_BUCKET '
+                    'environment variable.'
+                )
+            self._aws_backend_bucket = os.environ.get('AWS_BACKEND_BUCKET', f'{account_number}-terraform-state')
+            self._aws_backend_key = f'{self._type}.tfstate'
 
     def _get_config(self):
         """
@@ -67,8 +76,7 @@ class EC2Manager:
         with open(self._config_file, 'r') as config_file:
             return yaml.safe_load(config_file)
 
-    @staticmethod
-    def _get_json_from_s3(bucket_name, file_name):
+    def _get_json_from_s3(self, bucket_name, file_name):
         """
         Gets a json file from an s3 bucket.
 
@@ -79,8 +87,12 @@ class EC2Manager:
         """
         s3 = boto3.resource('s3')
         content_object = s3.Object(bucket_name, file_name)
-        file_content = content_object.get()['Body'].read().decode('utf-8')
-        return json.loads(file_content)
+        try:
+            file_content = content_object.get()['Body'].read().decode('utf-8')
+            return json.loads(file_content)
+        except Exception:
+            self._logger.warning('backend state file does not exist.')
+            return {}
 
     @staticmethod
     def _get_encoded_envs(encoded_string):
@@ -97,7 +109,7 @@ class EC2Manager:
                 if '=' in line}
 
     @staticmethod
-    def _encode_string(string):
+    def encode_string(string):
         """
         Encodes string to base64 string
         """
@@ -117,14 +129,14 @@ class EC2Manager:
         """
         Get the instance id by tag name:
 
-        :param str name: The name of the bot.
+        :param str name: The name of the instance.
         :param bool name: Whether to make one attempt to get the instance id.
         :param int attempts: The number of attempts made to get the instance id.
         :returns: Instance id.
         :rtype: str
         """
         if not try_once:
-            logging.info(f"{name} has been trying to get its instance id for {attempts} secs...")
+            self._logger.info(f"{name} has been trying to get its instance id for {attempts} secs...")
         instance_id = self.list_instances().get(name, {}).get('instance_id')
         if not instance_id and attempts < self._max_attempts and not try_once:
             time.sleep(5)
@@ -137,31 +149,23 @@ class EC2Manager:
 
         :return dict: A dictionary of output values.
         """
-        state_file_path = os.path.join(self._terraform_directory, '.terraform', 'terraform.tfstate')
-
-        if os.path.exists(state_file_path):
-            with open(state_file_path, 'r') as state_file:
-                data = json.load(state_file)
-                bucket_name = data['backend']['config']['bucket']
-                file_name = data['backend']['config']['key']
-                state_data = self._get_json_from_s3(bucket_name, file_name)
-                return state_data.get('outputs', {})
-        return {}
+        state_data = self._get_json_from_s3(self._aws_backend_bucket, self._aws_backend_key)
+        return state_data.get('outputs', {})
 
     def _get_instance_names(self):
         """
-        Gets a list of bot names whose instances should be created.
+        Gets a list of instance names whose instances should be created.
 
-        :return list[str]: A list of bot names.
+        :return list[str]: A list of instance names.
         """
         instances = self._config.get('instances') or {}
         return list(instances.keys())
 
     def _get_instance_repo_updates(self):
         """
-        Gets a list of bot names whose repos should be updated.
+        Gets a list of instance names whose repos should be updated.
 
-        :return list[str]: A list of bot names.
+        :return list[str]: A list of instance names.
         """
         instances = self._config.get('instances') or {}
         return [key for key, value in instances.items() if value.get('update')]
@@ -172,8 +176,8 @@ class EC2Manager:
 
         :return datetime: A UTC datetime object.
         """
-        if self._commit != 'master':
-            repo = self._github_client.get_repo(full_name_or_id=f'{self._github_user}/{self._repo}')
+        if self._commit not in ['master', 'main']:
+            repo = self.get_repo()
             commit_time = repo.get_commit(self._commit).last_modified
             return datetime.strptime(commit_time, '%a, %d %b %Y %H:%M:%S %Z').astimezone(pytz.UTC)
 
@@ -181,7 +185,7 @@ class EC2Manager:
         """
         Gets all the instance's environment variables.
 
-        :return dict: A dictionary of bot names and their environment variables.
+        :return dict: A dictionary of instance names and their environment variables.
         """
         instances = self._config['instances']
         default_ports = [{'from_port': 2375, 'to_port': 2375, 'protocol': 'tcp'}] if self._expose_docker_daemons else []
@@ -193,7 +197,7 @@ class EC2Manager:
             },
             'volume_size': instances[name].get('volume_size', 8),
             'instance_type': instances[name].get('instance_type', 't4g.nano'),
-            'working_directory': instances[name].get('working_directory', ''),
+            'working_directory': instances[name].get('working_directory', 'compose'),
             'ports': instances[name].get('ports', []) + default_ports,
             'envs': {
                 'NAME': name,
@@ -210,19 +214,20 @@ class EC2Manager:
         """
         Check for the given status on the ec2 instance by id.
 
-        :param str name: The name of the bot.
+        :param str name: The name of the instance.
         :param str status: The name of the ec2 instance status check to wait for.
         """
-        logging.info(f"{name} instance {self._instance_id} is checking if its status is {status.split('_')[-1]}")
-        if self._instance_id:
+        instance_id = self._get_instance_id(name)
+        self._logger.info(f"{name} instance {instance_id} is checking if its status is {status.split('_')[-1]}")
+        if instance_id:
             try:
                 waiter = self._ec2_client.get_waiter(status)
-                waiter.wait(InstanceIds=[self._instance_id])
-                logging.info(f"{name} instance {self._instance_id} is {status.split('_')[-1]}")
+                waiter.wait(InstanceIds=[instance_id])
+                self._logger.info(f"{name} instance {instance_id} is {status.split('_')[-1]}")
             except WaiterError:
-                logging.error(f"{name} status check of instance {self._instance_id} failed")
+                logging.error(f"{name} status check of instance {instance_id} failed")
         else:
-            logging.info(f"{name} instance does not exist")
+            self._logger.info(f"{name} instance does not exist")
 
     def _create_backend_bucket(self):
         """
@@ -230,7 +235,7 @@ class EC2Manager:
         """
         buckets = [item['Name'] for item in self._s3_client.list_buckets().get('Buckets', [])]
         if self._aws_backend_bucket not in buckets:
-            logging.info(f'Creating Terraform backend...')
+            self._logger.info(f'Creating Terraform backend...')
             self._s3_client.create_bucket(
                 ACL='private',
                 Bucket=self._aws_backend_bucket
@@ -243,14 +248,14 @@ class EC2Manager:
         Runs terraform init.
         """
         self._create_backend_bucket()
-        logging.info(f'Initializing Terraform backend...')
-        process = subprocess.run(
-            [
-                'terraform',
-                'init',
-                f'-backend-config="key={self._type}.tfstate"',
-                f'-backend-config="bucket={self._aws_backend_bucket}"'
-            ],
+        self._logger.info(f'Initializing Terraform backend...')
+        process = subprocess.run([
+            'terraform',
+            'init',
+            f'-backend-config=key={self._aws_backend_key}',
+            f'-backend-config=bucket={self._aws_backend_bucket}',
+            f'-backend-config=region={self._aws_default_region}'
+        ],
             env=os.environ,
             cwd=self._terraform_directory
         )
@@ -263,8 +268,8 @@ class EC2Manager:
         """
         outputs = self._get_terraform_outputs()
         if outputs.get('instance_names', {}).get('value') != self._instance_names:
-            logging.info(f'Terraform instances...')
-            instances = self._encode_string(json.dumps(self._get_instance_data()))
+            self._logger.info(f'Terraform instances...')
+            instances = self.encode_string(json.dumps(self._get_instance_data()))
 
             process = subprocess.run(
                 [
@@ -292,13 +297,65 @@ class EC2Manager:
             if process.returncode != 0:
                 raise RuntimeError('Failed to apply terraform')
 
+    def _call_on_all(self, callable_instance, items, extra_args=None, **kwargs):
+        """
+        Helper method for calling a given method on all provided.
+
+        :param callable callable_instance: A callable.
+        :param list items: A list of items to call.
+        :param list extra_args: A list of extra arguments.
+        :param bool join: Whether or not to join the threads.
+        :return list[Thread]: A list of threads.
+        """
+        if not extra_args:
+            extra_args = []
+
+        if type(items) != dict:
+            if type(items) != list:
+                items = list(items)
+            new_items = {item: {} for item in items}
+        else:
+            new_items = items
+
+        threads = []
+        # begin call on each item in a separate thread
+        for name, data in new_items.items():
+            args = []
+            args.insert(0, name)
+            if data:
+                args.insert(1, data)
+
+            thread = threading.Thread(
+                name=name,
+                target=callable_instance,
+                args=args+extra_args,
+                kwargs=kwargs
+            )
+            thread.start()
+            threads.append(thread)
+
+        # join all threads which waits for each call to finish
+        if kwargs.get('join', True):
+            for thread in threads:
+                thread.join()
+            return []
+        return threads
+
     def _await_instances(self):
         """
         Waits till all ec2 instances have a status of ok.
         """
-        for name in self._get_instance_names():
-            self._instance_id = self._get_instance_id(name)
-            self._status_check(name, 'instance_status_ok')
+        self._call_on_all(
+            callable_instance=self._status_check,
+            items=self._get_instance_names(),
+            extra_args=['instance_status_ok']
+        )
+
+    def get_repo(self):
+        """
+        Gets the repo instance.
+        """
+        return self._github_client.get_repo(full_name_or_id=f'{self._github_user}/{self._repo}')
 
     def list_instances(self):
         """
@@ -387,62 +444,91 @@ class EC2Manager:
         # wait for each instance status to be ok
         self._await_instances()
 
+    def _update_repo(self, name):
+        """
+        Updates a single instance repo.
+        """
+        if self._repo not in self.run_command(name, ['ls'], print_output=False):
+            self._logger.info(f'{name} cloning {self._repo}')
+            self.run_command(name, [f'git clone {self._repo_url}'])
+
+        self._logger.info(f'{name} checking out {self._commit}')
+        self.run_command(name, [
+            f'cd ./{self._repo}',
+            'git fetch',
+            'git pull origin master',
+            f'git -c advice.detachedHead=false checkout {self._commit}'
+        ])
+
+    def _stop(self, name, data):
+        """
+        Stops a single docker container if it is already running.
+        """
+        # check if docker is already running
+        result = list(filter(None, self.run_command(name, ['docker ps'], print_output=False).split('\n')))
+        if len(result) > 1:
+            # then compose down
+            self._logger.info(f'{name} stopping...')
+            working_directory = self._get_working_directory(data['working_directory'])
+            self.run_command(name, [
+                f'cd {working_directory} ',
+                data['commands']['stop']
+            ])
+
+    def _start(self, name, data):
+        """
+        Runs start command on a single instance.
+        """
+        self._logger.info(f'{name} starting...')
+        working_directory = self._get_working_directory(data['working_directory'])
+        self.run_command(name, [
+            f'cd {working_directory} ',
+            *[f'export {key}={value}' for key, value in data['envs'].items()],
+            data['commands']['start'],
+            *[f'unset {key}' for key in data['envs'].keys()]
+        ])
+
     def update_repos(self):
         """
         Updates the code in the instance repos.
         """
-        for name in self._instance_repo_updates:
-            if self._repo not in self.run_command(name, ['ls'], print_output=False):
-                logging.info(f'{name} cloning {self._repo}')
-                self.run_command(name, [f'git clone {self._repo_url}'])
-
-            logging.info(f'{name} checking out {self._commit}')
-            self.run_command(name, [
-                f'cd ./{self._repo}',
-                'git fetch',
-                f'git checkout {self._commit}',
-                'git pull',
-            ])
+        self._call_on_all(
+            callable_instance=self._update_repo,
+            items=self._instance_repo_updates,
+        )
 
     def stop(self):
         """
         Runs stop if a docker container is already running.
         """
-        for name, data in self._get_instance_data().items():
-            # check if docker is already running
-            result = list(filter(None, self.run_command(name, ['docker ps'], print_output=False).split('\n')))
-            if len(result) > 1:
-                # then compose down
-                logging.info(f'{name} stopping...')
-                working_directory = self._get_working_directory(data['working_directory'])
-                self.run_command(name, [
-                    f'cd {working_directory} ',
-                    data['commands']['stop']
-                ])
+        self._call_on_all(
+            callable_instance=self._stop,
+            items=self._get_instance_data(),
+        )
 
     def start(self):
         """
         Runs start command on the instances.
         """
-        for name, data in self._get_instance_data().items():
-            logging.info(f'{name} starting...')
-            working_directory = self._get_working_directory(data['working_directory'])
-            self.run_command(name, [
-                f'cd {working_directory} ',
-                *[f'export {key}={value}' for key, value in data['envs'].items()],
-                data['commands']['start'],
-                *[f'unset {key}' for key in data['envs'].keys()]
-            ])
+        self._call_on_all(
+            callable_instance=self._start,
+            items=self._get_instance_data(),
+        )
 
     def init(self):
         """
         Initializes the project
         """
+        destination = os.path.join(os.getcwd())
         if not os.path.exists(self._terraform_directory):
             shutil.copytree(
                 src=self._template_directory,
-                dst=os.path.join(os.getcwd(), 'ec2_manager')
+                dst=destination,
+                dirs_exist_ok=True
             )
+            self._logger.info(f'Initialized project at "{destination}"')
+        else:
+            self._logger.info(f'Project is already initialized project at "{destination}"')
 
     def apply(self):
         """
@@ -452,32 +538,3 @@ class EC2Manager:
         self.update_repos()
         self.stop()
         self.start()
-
-
-@click.group()
-def cli():
-    pass
-
-
-@click.command()
-def init():
-    ec2_manager = EC2Manager()
-    ec2_manager.init()
-
-
-@click.command()
-@click.option('--config', default='config.yaml', help='Path to config file')
-@click.option('--repo', default=os.environ.get('REPO'), help='The name of the repo i.e. "Hello-World"', required=True)
-@click.option('--github-username', default=os.environ.get('GITHUB_USERNAME'), help='Your github username i.e. "octocat"', required=True)
-@click.option('--github-token', default=os.environ.get('GITHUB_TOKEN'), help='A github personal access token that has permission to pull the repo', required=True)
-def apply(**kwargs):
-    ec2_manager = EC2Manager(**kwargs)
-    ec2_manager.apply()
-
-
-def main():
-    cli.add_command(init)
-    cli.add_command(apply)
-    cli()
-
-sys.exit(main())
\ No newline at end of file
